#include <iostream>
#include <cstring>
#include <cstdio>
#include "casegen.hh"

#include <unistd.h>
#include <map>
#include <set>

static unsigned LabelCounter = 0;
static int rnd;

class x86case: public CaseGenerator
{
    bool DeadCode;
    bool ChainCompare;
    CaseValue LastCompare;
    std::string LastCompareType;
    int bp_state;
public:
    virtual CaseValue GetMinValue()     const { return 0; }
    virtual CaseValue GetMaxValue()     const { return 0xFF; }
    virtual CaseValue GetDefaultCase()  const { return -1; }

    x86case() : DeadCode(false), ChainCompare(false), LastCompare(-1), bp_state(-1)
    {
    }

    void EmitCompareJump(CaseValue value, const std::string& label, const char* comparetype)
    {
        if(!ChainCompare || value != LastCompare)
        {
            std::cout << "\tcmp al, ";
            if(value == '\'' || value == '\\' || value < 32 || value > 126)
                std::cout << value;
            else
                std::cout << '\'' << char(value) << '\'';
            std::cout << "\n";
            LastCompare = value;
            ChainCompare = true;
        }
        std::cout << "\t" << comparetype << " " << label << "\n";
        LastCompareType = comparetype;
    }

    virtual void EmitJump(const std::string& label)
    {
        if(DeadCode) return;
        std::cout << "\tjmp " << label << "\n";
        ChainCompare = false;
        DeadCode = true;
        bp_state = -1;
    }
    virtual void EmitSubtract(CaseValue value)
    {
        std::cout << "\tsub al, ";
        if(value == '\'' || value == '\\' || value < 32 || value > 126)
            std::cout << value;
        else
            std::cout << '\'' << char(value) << '\'';
        std::cout << "\n";
        std::cout << std::flush;

        ChainCompare = true;
        LastCompare  = 0;
        LastCompareType = "je";
        bp_state = -1;
    }
    virtual const std::string EmitBlock()
    {
        char Buf[64];
        std::sprintf(Buf, "L%d_%u", rnd, LabelCounter++);
        return Buf;
    }
    virtual void EmitEndBlock(const std::string& label)
    {
        std::cout << label << ":\n";
        ChainCompare = false;
        DeadCode = false;
        bp_state = -1;
    }
    virtual void EmitJumpTable(const std::vector<std::string>& table)
    {
        char Buf[64];
        std::sprintf(Buf, "T%d_%u", rnd, LabelCounter++);
        std::string TableName = Buf;

        std::cout <<
        //    "\tmovzx eax,al\n"
        //    "\tjmp [cs:" << TableName << " + eax*2]\n"
            "\tmovzx bp,al\n"
            "\tshl bp,1\n"
            "\tjmp [bp + " << TableName << "]\n"
        ;
        std::cout << TableName << ":\n";
        std::string line;
        for(size_t a=0; a<table.size(); ++a)
        {
            if(line.empty()) line += "\tdw "; else line += ", ";
            line += table[a];
            if(line.size() > 60) { std::cout << line << "\n"; line.clear(); }
        }
        if(!line.empty()) { std::cout << line << "\n"; }
        ChainCompare = false;
        bp_state = -1;
        std::cout << std::flush;
    }
    virtual void EmitCompareEQ(CaseValue value, const std::string& target)
    {
        EmitCompareJump(value, target, "je");
    }
    virtual void EmitCompareNE(CaseValue value, const std::string& target)
    {
        EmitCompareJump(value, target, "jne");
    }
    virtual void EmitCompareLT(CaseValue value, const std::string& target)
    {
        if(value > GetMaxValue())
            EmitJump(target);
        else if(value == GetMinValue())
            { /* nothing */ }
        else
        {
            if(ChainCompare && value == LastCompare) goto Yay;
            if(value == GetMinValue()+1)
                EmitCompareEQ(value-1, target);
            else if(Desirable(value-1))
                EmitCompareLE(value-1, target);
            else
                Yay: EmitCompareJump(value, target, "jb");
        }
    }
    virtual void EmitCompareGT(CaseValue value, const std::string& target)
    {
        if(value < GetMinValue())
            EmitJump(target);
        else if(value == GetMaxValue())
            { /* nothing */ }
        else if(ChainCompare && value == LastCompare && LastCompareType == "je")
            EmitCompareGE(value, target);
        else
        {
            if(ChainCompare && value == LastCompare) goto Yay;
            if(value == GetMaxValue()-1)
                EmitCompareEQ(value+1, target);
            else if(Desirable(value+1))
                EmitCompareGE(value+1, target);
            else
                Yay: EmitCompareJump(value, target, "ja");
        }
    }
    virtual void EmitCompareGE(CaseValue value, const std::string& target)
    {
        if(value == GetMinValue())
            EmitJump(target);
        else if(value == GetMaxValue())
            EmitCompareEQ(value, target);
        else
        {
            if(ChainCompare && value == LastCompare) goto Yay;
            if(Desirable(value-1))
                EmitCompareGT(value-1, target);
            else
                Yay: EmitCompareJump(value, target, "jae");
        }
    }
    virtual void EmitCompareLE(CaseValue value, const std::string& target)
    {
        if(value == GetMaxValue())
            EmitJump(target);
        else if(value == GetMinValue())
            EmitCompareEQ(value, target);
        else
        {
            if(ChainCompare && value == LastCompare) goto Yay;
            if(Desirable(value+1))
                EmitCompareLT(value+1, target);
            else
                Yay: EmitCompareJump(value, target, "jbe");
        }
    }
    virtual void EmitTestBits16
        (unsigned bitmask, const std::string& target, bool assume_capped)
    {
        std::string skip;
        if(!assume_capped)
        {
            skip = EmitBlock();
            EmitCompareGE(16, skip);
        }

        ChainCompare = false;
        // This requires a 32-byte lookup table, but preserves CX
        if(bp_state == -1)
        {
            std::cout << "\tmovzx bp,al\n";
            bp_state = 1;
        }
        if(bp_state == 1)
        {
            std::cout << "\tshl bp,1\n";
            bp_state = 2;
        }
        std::cout <<
            "\ttest word [bp + powers_of_two], 0x" << std::hex << bitmask << std::dec << "\n"
            "\tjnz " << target <<"\n";

        if(!assume_capped)
        {
            std::cout << skip << ":\n";
            bp_state = -1;
        }

        std::cout << std::flush;
    }

    virtual void EmitTestBits8
        (unsigned bitmask, const std::string& target, bool assume_capped)
    {
        std::string skip;
        if(!assume_capped)
        {
            skip = EmitBlock();
            EmitCompareGE(8, skip);
        }

        ChainCompare = false;
        // This requires a 8-byte lookup table, but preserves CX
        if(bp_state == 2)
        {
            std::cout <<
                "\ttest byte [bp + powers_of_two], 0x" << std::hex << bitmask << std::dec << "\n"
                "\tjnz " << target <<"\n";
        }
        else
        {
            if(bp_state != 1)
            {
                std::cout << "\tmovzx bp,al\n";
                bp_state = 1;
            }
            std::cout <<
                "\ttest byte [bp + powers_of_two8], 0x" << std::hex << bitmask << std::dec << "\n"
                "\tjnz " << target <<"\n";
        }

        if(!assume_capped)
        {
            std::cout << skip << ":\n";
            bp_state = -1;
        }

        std::cout << std::flush;
    }

    virtual void EmitTestMask8(unsigned bitmask, const std::string& target)
    {
        std::cout <<
            "\ttest al, 0x" << std::hex << bitmask << std::dec << "\n"
            "\tjz " << target << "\n";
        std::cout << std::flush;
        ChainCompare = false;
    }

private:
    bool Desirable(CaseValue v)
    {
        if(v < 0 || v > 255) return false;
        if(ChainCompare && v == LastCompare) return true;
        switch(v)
        {
            case '0':
            case '9':
            case 'A':
            case 'F':
            case 'Z':
            case 'a':
            case 'f':
            case 'z':
                return true;
        }
        return false;
    }
};

int main()
{
    rnd = getpid();


    std::string defaultlabel;
    std::map<std::string/*target*/, std::set<long>/*values*/> data;

    char Buf[512];
    while((fgets(Buf,512, stdin)))
    {
        std::strtok(Buf, "\n");

        if(defaultlabel.empty())
        {
            char Buf2[512];
            if(std::sscanf(Buf, "%s %d", Buf2, &rnd) == 2)
                defaultlabel = Buf2;
            else
                defaultlabel = Buf;
            continue;
        }

        long value=-1;
        std::sscanf(Buf, "%ld %s", &value, Buf);
        std::set<long>& values = data[Buf];
        if(value >= 0 && value <= 255)
            values.insert(value);
    }

    CaseItemList list;
    for(std::map<std::string, std::set<long> >::const_iterator
        i = data.begin(); i != data.end(); ++i)
    {
        CaseItem tmp;
        tmp.target = i->first;
        tmp.values = i->second;
        list.push_back(tmp);
    }

    x86case engine;
    engine.Generate(list, defaultlabel);

    return 0;
}
